
#include "VKBaseData.hh"

#include <TParameter.h>

using namespace std;


VKBaseData::VKBaseData() 
 : fAddRes(false)
{
    // coefficients of the energy-resolution function

    //Old parameters:
    //fP0  = 30.44;
    //fP1  = 17.30;
    //fTauRel = 0;

    fP0  = 0.0;
    fP1  = 5.5;
    fTauRel = 1.93;
    
    double emin = -3000;
    double emax = 100;
    
    // experimental resolution function
    if (fTauRel == 0) { // Gaussian   
        fRes = new TF1("fRes", "TMath::Gaus(x, [0], [1], true)", emin, emax);
    }
    else { // Gaussian with low-energy tail
        fRes = new TF1("fRes", "1./[2] * TMath::Exp((x-[0])/[2] + 0.5 * TMath::Power([1]/[2], 2)) * TMath::Erfc(1./sqrt(2) * ((x-[0])/[1] + [1]/[2]))", emin, emax);
    }
}


VKBaseData::~VKBaseData() 
{
    // make sure ROOT file has been closed
    if (fOutputFile != NULL && fOutputFile->IsOpen()) fOutputFile->Close();

    // delete pointers
    delete fOutputFile;
    delete fOutputTree;
}


void VKBaseData::CreateOutput(string filename) 
{
    cout << " Creating output file: " << filename << " ..." << endl;

    // random number generator
    fRNG = new TRandom3();

    // create tree
    fOutputFile = new TFile(filename.c_str(), "RECREATE");
    fOutputTree = new TTree("Detector", "Detector data"); 
    CreateOutputBranches(fOutputTree);
    
    // write normalization
    TParameter<double>* pNorm = new TParameter<double>("NORMALIZATION", this->GetNormalization());
    fOutputTree -> GetUserInfo() -> Add(pNorm);
    cout << " Normalization: " << this->GetNormalization() << endl;
    
    // write area
    TParameter<double>* pArea = new TParameter<double>("SOURCE_AREA/cm2", this->GetArea());
    fOutputTree -> GetUserInfo() -> Add(pArea);    
    
    // loop over events and fill output tree
    for (size_t i=0; i<this->GetEntries(); i++) {

        // get event data
        const auto& ev = this -> GetEvent(i);
        _ESout = ev.ES;
        _EVout = ev.EV;
        _PSout = ev.PS;
        _PVout = ev.PV;
        _EF2out = ev.EF2;
        _ES2out = ev.ES2;
        _EV2out = ev.EV2;
        
        // sample experimental resolution
        SampleResolution(_ESout);
        SampleResolution(_EVout);
        SampleResolution(_EF2out);
        SampleResolution(_ES2out);
        SampleResolution(_EV2out);

        // fill tree
        fOutputTree -> Fill();
    }

    // write tree to file
    fOutputTree -> Write();
    
    // close output file
    fOutputFile -> Close();
    
    cout << " Done!" << endl;
}


void VKBaseData::CreateOutputBranches(TTree * t) 
{
    t->Branch("PhotonsSignal", &_PSout, "_PSout/i");
    t->Branch("PhotonsVeto"  , &_PVout  , "_PVout/i");

    if (this->HasV2Branches()) {
        t->Branch("EDepV2Front", &_EF2out, "_EF2out/F");
        t->Branch("EDepV2Signal", &_ES2out, "_ES2out/F");
        t->Branch("EDepV2Veto"  , &_EV2out,   "_EV2out/F");
    }
    else {
        t->Branch("EDepSignal", &_ESout, "_ESout/F");
        t->Branch("EDepVeto"  , &_EVout,   "_EVout/F");
    }
}


void VKBaseData::SampleResolution(Float_t& e)
{
    if (!fAddRes) return;

    double fwhm = sqrt(pow(fP0,2) + fP1 * e);
    double sigma = fwhm / 2.35;
    double tau = sigma * fTauRel;

    // set parameter values
    fRes->SetParameter(0, 0);
    fRes->SetParameter(1, sigma);
    if (fTauRel > 0) {
        fRes->SetParameter(0, tau); // shift function so mean is at zero
        fRes->SetParameter(2, tau);
    }
    
    if (fwhm > 50 && e > 50) {
        e += fRes->GetRandom();
    }
    
    if (e<=0) e = 0;
}
