#ifndef VKMaterials_h
#define VKMaterials_h 1

#include "globals.hh"
#include "G4Element.hh"
#include "G4Material.hh"
#include "G4NistManager.hh"

/**
 * Constructs all materials including specification of optical properties.
 * Implemented as a static class.
 * 
 * Adapted from: 
 * https://github.com/murffer/DetectorSim/blob/master/ScintillationSlab/include/Materials.hh
 */
class VKMaterials {

  public:
    ~VKMaterials();
    static VKMaterials* GetInstance();
    G4Material* GetMaterial(const G4String);

  private:
    VKMaterials();

    /**
     * Creates the materials
     */
    void CreateMaterials();

    /* 
     * Sets the optical properities of EJ-200 Scintillator.
     *
     * The EJ-200 scintillator material is G4_PLASTIC_SC_VINYLTOLUENE.
     * Data sources: 
     *  - http://fcomp.ist.utl.pt/LRC.web/equipamento/EJ200-datasheet.pdf
     *  - http://hypernews.slac.stanford.edu/HyperNews/geant4/get/AUX/2014/07/18/09.40-37419-DetectorConstruction.cc
     *  - https://github.com/murffer/DetectorSim/blob/master/ScintillationSlab/src/Material.cc
     */ 
    void SetOpticalPropertiesEJ200();

    /**
     * Sets the optical properities of PMMA.
     *
     * The PMMA material is G4_PLEXIGLASS.
     * Data Sources:
     *  - Refractive Index is from S.N. Kasarova (Table 1) doi:10.1016/j.optmat.2006.07.010
     *  - Absorbance is based on formula on pg. 139 of Urffer Notebook II, and on the transmission data from Eljen
     */
    void SetOpticalPropertiesPMMA();

    /**
     * Sets the optical properties of Teflon
     *
     * Teflon is also Polytetrafluoroethylene, which could be flurinated polyethene
     * Sources of the Teflon data:
     *  - Index of Reflection: DOI 10.1117/1.2965541
     *  - Absorbition length: GEANT4 lXe example
     */
    void SetOpticalPropertiesTeflon();

    /**
     * Sets the optical properties of Glass
     *
     * Source: GEANT4 lXe example
     */
    void SetOpticalPropertiesGlass();

    static VKMaterials* instance;   /* Materials Instance */
    G4NistManager* nistMan;       /* NIST Material Manager */

    G4Element *fH;
    G4Element *fC;
    G4Element *fN;
    G4Element *fO;
    G4Element *fAl;
    G4Element *fSi;
    G4Element *fAr;
    G4Element *fFe;
    G4Element *fCu;
    G4Element *fZn;
    G4Element *fBr;
    G4Element *fLa; 
    G4Element *fCe;
    G4Element *fPb;
    
    G4Material *fLabVacuum;
    G4Material *fSilicon;
    G4Material *fCopper;
    G4Material *fIron;
    G4Material *fBrass;
    G4Material *fCarbon;
    G4Material *fAluminium;
    G4Material *fArgon;
    G4Material *fNitrogen;
    G4Material *fOxygen;
    G4Material *fEJ200;
    G4Material *fPMMA;
    G4Material *fTeflon;
    G4Material *fGlass;
    G4Material *fLead;
};

#endif
