//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#ifndef VKPrimaryGeneratorAction_h
#define VKPrimaryGeneratorAction_h 1

#include "G4VUserPrimaryGeneratorAction.hh"
#include "VKBetaSpectrum.hh"

//class G4ParticleGun;
class G4Event;
class G4GeneralParticleSource;
class TF1;
class VKPrimaryGeneratorMessenger;


class VKPrimaryGeneratorAction : public G4VUserPrimaryGeneratorAction
{
  public:

    VKPrimaryGeneratorAction();
    virtual ~VKPrimaryGeneratorAction();
 
    virtual void GeneratePrimaries(G4Event* anEvent);
    
    // Enable/disable CR muon angular distribution
    void SetCRMuonAngDistOn(G4bool b);
    
    // Set value of exponent in CR muon angular distribution formula
    void SetCRMuonAngDistPower(float power);

    // Set min and max angles for sampling of CR muons
    void SetCRMuonMinZenithAngle(G4double angle);
    void SetCRMuonMaxZenithAngle(G4double angle);
    
    // Set size and orientation of source disk
    void SetSourceInnerDiameter(G4double d);
    void SetSourceOuterDiameter(G4double d);
    void SetSourceNormal(const G4ThreeVector& n);

    // Enable/disable beta-decay energy spectrum (allowed shape)
    void SetBetaSpectrumOn(G4bool b);

    // Enable/disable sampling of disk-shaped source
    void SetDiskShapedSourceOn(G4bool b);
    
    // Set end-point energy and min and max energy for sampling
    void SetBetaEndPointEnergy(G4double endPoint);
    void SetBetaMinEnergy(G4double emin);
    void SetBetaMaxEnergy(G4double emax);
    
    // Set coefficients of beta shape function
    void SetBetaShapeFuncA1(double a1);
    void SetBetaShapeFuncA2(double a2);
    void SetBetaShapeFuncA3(double a3);
    void SetBetaShapeFuncA4(double a4);
    void SetBetaShapeFuncA5(double a5);
    void SetBetaShapeFuncA6(double a6);
    
    // Get coverage of sampling
    double GetCoverage() const;
    
    // Get centre position of source
    inline const G4ThreeVector& GetCentrePosition() const {return fPosCentre;};

    // Get coverage of sampling
    double GetSourceArea() const;
    
    // Get particle name
    G4String GetParticleName() const;
    
    // Get enable/disable options
    inline G4bool GetCRMuonAngDistOn() const {return fCRMuonAngDistOn;};
    inline G4bool GetBetaSpectrumOn() const {return fBetaSpectrumOn;};
    inline G4bool GetDiskShapedSourceOn() const {return fDiskShapedSourceOn;};

  private:

    void InitCRMuonAngDist();
    void InitRadialDist();
    void GetOrthogonalVectors(const G4ThreeVector& n, G4ThreeVector& o1, G4ThreeVector& o2);
    void InitBetaSpectrum();
    double GetBetaCoverage() const;
    double GetCRMuonCoverage() const;

    VKPrimaryGeneratorMessenger * primGen_msg;
    
    G4GeneralParticleSource* fGPSource;
    G4bool fGPSourceOn;
    
    G4ThreeVector fDirection;
    G4ThreeVector fPosCentre;
    G4ThreeVector fPos;
    
    float fPower;
    double fCRMuonCoverage;
    G4bool fCRMuonAngDistOn;
    G4double fZenithMin;
    G4double fZenithMax;
    TF1 * fCRMuonZenithDist;
    TF1 * fCRMuonAzimuthalDist;
    
    G4double fElectronMass;
    G4double fBetaLowCut;
    G4double fBetaEKinMin;
    G4double fBetaEKinMax;
    G4double fBetaEKinEndPoint;
    double fBetaCoverage;
    VKBetaSpectrum * fBetaSpectrum;
    TF1 * fBetaSpectrumFunc;
    TF1 * shapeFunc;
    G4bool fBetaSpectrumOn;
    double shapeFuncCoeff[6];
    
    G4bool fDiskShapedSourceOn;
    G4double fSourceInnerDiameter;
    G4double fSourceOuterDiameter;
    G4ThreeVector fSourceNormal;
    G4ThreeVector fOrtho1, fOrtho2;
    TF1 * fRadialDist;
    
};

#endif
