
#include "VKBetaSpectrum.hh"

#include "G4Electron.hh"
#include "G4SystemOfUnits.hh"
#include "globals.hh"

#include "TMath.h"

#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_result.h>
#include <gsl/gsl_sf_dilog.h>

//TF1 * shapeFunc = new TF1("shapeFunc", "0.9674 + 0.01186*x + 0.000333*pow(x,2) - 0.0021/x", 0.5, 1E3);


VKBetaSpectrum::VKBetaSpectrum(int Z_daughter, int A, G4double endPointKinEnergy, TF1 * shapeFunc)
 : Z(Z_daughter), A(A), shapeFunc(shapeFunc)
{
    me = G4Electron::ElectronDefinition()->GetPDGMass() / MeV;

    double T0 = endPointKinEnergy / MeV;
    W0 = (T0 + me) / me;

    alpha = 1 / 137.036; // fine-structure constant
    gamma = sqrt(1 - pow(alpha*Z, 2));
    R = 1.2 * pow(A, 1/3.) / 197 * me; // nuclear radius in units of me^-1
    d1 = R / 3;
    M = 1837; // nucleon mass in units of me
}


VKBetaSpectrum::~VKBetaSpectrum() 
{
    // do nothing
}


// Wilkinson and Marrs, Nucl. Instrum. Meth. 5 (1972) 505-508
double VKBetaSpectrum::Evaluate(double *x, double *par)
{
    double W = (x[0] + me) / me;
    double p = sqrt(pow(W, 2) - 1);
    double pnu = W0 - W;
    double y = alpha * Z * W / p;
    double Wbar = W + 3 * alpha * Z / (2*R);
    double pbar = sqrt(pow(Wbar, 2) - 1);
    double H2 = -pow(pbar*R, 2) / 6;
    double N1 = pnu * R / 3;
    double N2 = -pow(pnu * R, 2) / 6;
    double D1 = Wbar * R / 3;
    
    double mu1 = 1; // (See Wilkinson and Macefield 1974, p. 65)

    // Evaluate complex gamma function
    gsl_sf_result lnr, arg;
    gsl_sf_lngamma_complex_e(gamma, y, &lnr, &arg);

    // fermi function F(Z,W)
    double F = 2*(1+gamma)*pow(2*p*R, -2*(1-gamma)) 
        * pow(TMath::Exp(lnr.val), 2) * pow(gsl_sf_gamma(2*gamma+1), -2) 
        * TMath::Exp(TMath::Pi() * y);
    double F0 = 2 * F / (1 + gamma);    

    // L0
    double a = 0.206;
    double b = -1.039;
    double c = -0.0058;
    double d = -0.54;
    double L0 = 1 + a * pow(alpha*Z, 2) 
        + b * W * R * alpha * Z
        + c * pow(p * R, 2)
        + d * R * alpha * Z / W;   
    
    // C
    double C = 1 + (6/5) * (H2 + N1 * D1 / 3 + N2)
        - 2 * mu1 * gamma * N1 * d1 / (5 * W);
    
    // bremssstrahlung factor
    double R = bremsstrahlungFactor(W, W0);
    
    // put it all together
    double f = R * p * W * pow(W0 - W, 2) * F0 * L0 * C;
    
    if (shapeFunc != nullptr) {
        double E = x[0] + me;
        f *= shapeFunc->Eval(E);
    }
    
    return f;
}


double VKBetaSpectrum::bremsstrahlungFactor(double W, double W0) {

    double p = sqrt(pow(W, 2) - 1);
    double beta = p / W;
    
    // spence function
    double x = 2 * beta / (1 + beta);
    double L = -gsl_sf_dilog(x);
    
    double g = 3 * TMath::Log(M) - 3/4 
        + 4 * (TMath::ATanH(beta)/beta - 1)
        * ((W0 - W)/(3*W) - 3/2 + TMath::Log(2*(W0-W)))
        + 4 / beta * L + 1 / beta * TMath::ATanH(beta)
        * (2 * (1 + pow(beta, 2)) + pow(W0/W - 1, 2) / 6 
        - 4 * TMath::ATanH(beta));
        
    return 1 + alpha / (2 * TMath::Pi()) * g;
}

// Sirlin, Phys. Rev. 164 (1967) 1767
