//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//
#include "globals.hh"

#include "VKMagnet.hh"
#include "VKUtil.hh"

#include "G4LogicalSkinSurface.hh"
#include "G4LogicalBorderSurface.hh"
#include "G4SystemOfUnits.hh"
#include "G4Cons.hh"


VKMagnet::VKMagnet(G4RotationMatrix *pRot,
                             const G4ThreeVector &tlate,
                             G4LogicalVolume *pMother_log,
                             G4bool pMany,
                             G4int pCopyNo,
                             const VKDetectorConstruction* pDetConstr)

 // Pass info to the G4PVPlacement constructor
 : G4PVPlacement(pRot, 
                 tlate,
                 CreateTemporaryLogicalVolume(), // <-- see VKUtil.cpp
                 "magnet", 
                 pMother_log, 
                 pMany, 
                 pCopyNo),
   fDetConstr(pDetConstr),              
   fMother_log(pMother_log)
{

    // default dimensions    
    SetDefaults();

    // List of materials
    fMaterials = VKMaterials::GetInstance();

    // Get materials
    G4Material* labVacuum = fMaterials -> GetMaterial("LABVACUUM");
    G4Material* copper    = fMaterials -> GetMaterial("COPPER");
    G4Material* iron      = fMaterials -> GetMaterial("IRON");
    G4Material* brass     = fMaterials -> GetMaterial("BRASS");
    G4Material* pmma      = fMaterials -> GetMaterial("G4_PLEXIGLASS");
    assert(copper != nullptr); // check that material exists
    assert(iron != nullptr);
    assert(brass != nullptr);
    assert(labVacuum != nullptr);
    assert(pmma != nullptr);
    
    // rotation matrices, translation and transformation
    G4RotationMatrix doNotRotate, rotate180DegAroundY; 
    doNotRotate.rotateX(0.);
    rotate180DegAroundY.rotateY(180.*deg);
    G4double displ;
    G4ThreeVector translation; 
    G4Transform3D transform;


    // -------------------------------------------- //
    //             (Imaginary) container            //
    // -------------------------------------------- //
    
    // *** All other volumes should fit inside this volume !

    // solid
    fContainer_sol = new G4Tubs("magnet_sol", 0., fContainerDiameter/2, 
                              fContainerLength/2, 0.*deg, 360.*deg);

    // logical
    fContainer_log = new G4LogicalVolume(fContainer_sol, labVacuum, "magnet_log");
    
    
    // ----------------------------------------------------------------- //
    //                              COILS                                //
    // ----------------------------------------------------------------- //

    // the coils are modelled as perfect tubes
    G4double startAngle, spanningAngle;
    G4Tubs* coil_sol = new G4Tubs("coil_sol", fCoilInnerDiameter/2, fCoilOuterDiameter/2,
                            fCoilWidth/2, startAngle=0.*deg, spanningAngle=360.*deg);

    // relative placement
    displ = fCoilSeparation + fCoilWidth;
    translation = G4ThreeVector(0, 0, displ); 
    transform = G4Transform3D(doNotRotate, translation);

    // create union
    fCoils_sol = new G4UnionSolid("coils_sol", coil_sol, coil_sol, transform);
                      
    fCoils_log = new G4LogicalVolume(fCoils_sol, copper, "coil_log");

    new G4PVPlacement(0, G4ThreeVector(0, 0, -displ/2), fCoils_log,
				      "coils", fContainer_log, false, 0, fCheckOverlaps);	


    // ----------------------------------------------------------------- //
    //                             IRON WALLS                            //
    // ----------------------------------------------------------------- //

    // Barrel
    G4Tubs* ironBarrel_sol = new G4Tubs("ironBarrel_sol",
                                        fIronBarrelInnerDiameter/2, 
                                        fIronBarrelInnerDiameter/2 + fIronBarrelThickness,
                                        fSpectrometerLength/2,
                                        startAngle=0.*deg, 
                                        spanningAngle=360.*deg);

    // Endcap
    G4Tubs* ironEndcap_sol = new G4Tubs("ironEndcap_sol",
                                        fIronEndcapHoleDiameter/2, 
                                        fIronBarrelInnerDiameter/2,
                                        fIronEndcapThickness/2,
                                        startAngle=0.*deg, 
                                        spanningAngle=360.*deg);
                 
    // Cone
    G4Cons* ironCone_sol = new G4Cons("ironCone_sol",
                                      fIronConeInnerDiameter/2,
                                      fIronConeOuterDiameterMax/2,
                                      fIronConeInnerDiameter/2,
                                      fIronConeOuterDiameterMin/2,
                                      fIronConeLength/2,
                                      startAngle=0.*deg, 
                                      spanningAngle=360.*deg);
                 
    // placement of cone relative to endcap
    displ       = (fIronConeLength + fIronEndcapThickness)/2;
    translation = G4ThreeVector(0, 0, displ); 
    transform   = G4Transform3D(doNotRotate, translation);

    // union of endcap and cone
    G4UnionSolid* endcapWithCone_sol = 
        new G4UnionSolid("ironWalls_sol", ironEndcap_sol, ironCone_sol, transform);

    // placement of (upstream) endcap relative to barrel
    displ       = -(fSpectrometerLength - fIronEndcapThickness)/2;
    translation = G4ThreeVector(0, 0, displ); 
    transform   = G4Transform3D(doNotRotate, translation);

    // union of barrel and (upstream) endcap+cone
    fIronWalls_sol = 
        new G4UnionSolid("ironWalls_sol", ironBarrel_sol, endcapWithCone_sol, transform);

    // placement of (downstream) endcap relative to barrel
    transform = G4Transform3D(rotate180DegAroundY, -translation);

    // union of barrel and (downstream) endcap+cone
    fIronWalls_sol = 
        new G4UnionSolid("ironWalls_sol", fIronWalls_sol, endcapWithCone_sol, transform);

    // logical volume
    fIronWalls_log = new G4LogicalVolume(fIronWalls_sol, iron, "ironWalls_log");  
  
    // physical volume
    new G4PVPlacement(0, G4ThreeVector(0, 0, 0), fIronWalls_log,
				      "ironWalls", fContainer_log, false, 0, fCheckOverlaps);


    // ----------------------------------------------------------------- //
    //                      BRASS WALLS (INNER)                          //
    // ----------------------------------------------------------------- //

    fBrassWalls_sol = new G4Tubs("brassWalls_sol",
                                 fBrassInnerDiameter/2, 
                                 fBrassInnerDiameter/2 + fBrassThickness,
                                 fBrassLength/2,
                                 startAngle=0*deg, 
                                 spanningAngle=360*deg);
  
    fBrassWalls_log = new G4LogicalVolume(fBrassWalls_sol, brass, "brassWalls_log");
  
    new G4PVPlacement(0, G4ThreeVector(0,0,0), fBrassWalls_log, 
                      "brassWalls", fContainer_log, false, 0, fCheckOverlaps);


/*-------------------------------------------------------------
    extra_sol = new G4Tubs("extra_sol",
                                 fBrassInnerDiameter/2 - 1.*cm, 
                                 fBrassInnerDiameter/2,
                                 3./2.*cm,
                                 startAngle=0*deg, 
                                 spanningAngle=360*deg);
    extra_log = new G4LogicalVolume(extra_sol, pmma, "extra_log");
    new G4PVPlacement(0, G4ThreeVector(0,0,0), extra_log, 
                      "extra", fContainer_log, false, 0, fCheckOverlaps);
--------------------------------------------------------------*/


    // visualization attributes
    VisAttributes();    


    // Set the logical volume
    this -> SetLogicalVolume(fContainer_log);
    
    // Move the detector so that 'tlate' coincides with the center of the magnet
    this -> SetTranslation(tlate);
    
    // Check for overlaps
    if (fCheckOverlaps) this -> CheckOverlaps();
}


void VKMagnet::SetDefaults(){

    fCheckOverlaps = fDetConstr -> GetCheckOverlaps();

    // coils
    fCoilInnerDiameter = 2*13.5*cm; 
    fCoilOuterDiameter = 2*24.0*cm;
    fCoilWidth         = 22.5*cm;
    fCoilSeparation    = 26.0*cm;     

    // iron walls (outer)
    fIronBarrelInnerDiameter   = 2*26.0*cm;
    fIronBarrelThickness       = 4.0*cm;
    fIronEndcapThickness       = fDetConstr -> GetIronEndcapThickness();
    fIronEndcapHoleDiameter    = fDetConstr -> GetIronEndcapHoleDiameter();
    fIronConeInnerDiameter     = fIronEndcapHoleDiameter; 
    fIronConeOuterDiameterMin  = 12.0*cm;
    fIronConeOuterDiameterMax  = 22.0*cm;
    fIronConeLength            = fDetConstr -> GetIronConeLength();
    fSpectrometerLength        = fDetConstr -> GetSpectrometerLength();
    
    // brass walls (inner)
    fBrassInnerDiameter = fDetConstr -> GetBoreDiameter();
    fBrassThickness     = 1.0*cm;
    fBrassLength        = fSpectrometerLength - 2*fIronEndcapThickness;
    
    // deduce container size
    // (entire beta detector should be able to fit inside so we make a little longer)
    fContainerDiameter = 2*(fIronBarrelInnerDiameter/2 + fIronBarrelThickness);
    fContainerLength   = fSpectrometerLength + 10*cm; 
}


void VKMagnet::VisAttributes() {

    // coils
    G4VisAttributes* coils_va = new G4VisAttributes(G4Colour(0.0,0.0,1.0)); // blue
    coils_va   -> SetVisibility(true);
    coils_va   -> SetForceWireframe(true);
    fCoils_log -> SetVisAttributes(coils_va);

    // iron walls
    G4VisAttributes* ironWalls_va = new G4VisAttributes(G4Colour(0.4,0.2,0.1));
    ironWalls_va   -> SetVisibility(true);
    ironWalls_va   -> SetForceWireframe(true);
    fIronWalls_log -> SetVisAttributes(ironWalls_va);
    
    // brass walls
    G4VisAttributes* brassWalls_va = new G4VisAttributes(G4Colour(0.0,1.0,0.0)); // green
    brassWalls_va   -> SetVisibility(true);
    brassWalls_va   -> SetForceWireframe(true);
    fBrassWalls_log -> SetVisAttributes(brassWalls_va);
}
