//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#include "VKPrimaryGeneratorAction.hh"
#include "VKPrimaryGeneratorMessenger.hh"

#include "G4Event.hh"
#include "G4ParticleTable.hh"
#include "G4ParticleDefinition.hh"
#include "G4SystemOfUnits.hh"
#include "globals.hh"
#include "G4GeneralParticleSource.hh"
#include "G4Electron.hh"

#include <TF1.h>
#include <TMath.h>


VKPrimaryGeneratorAction::VKPrimaryGeneratorAction()
 : fCRMuonAngDistOn(false),
   fBetaSpectrumOn(false),
   fDiskShapedSourceOn(false),
   fGPSourceOn(false),
   fPower(1.85),
   fCRMuonCoverage(1),
   fZenithMin(0*deg),
   fZenithMax(90*deg),
   fSourceInnerDiameter(0*mm),
   fSourceOuterDiameter(7*mm),
   fSourceNormal(G4ThreeVector(0,0,1)),
   fElectronMass(G4Electron::ElectronDefinition()->GetPDGMass()),
   fBetaLowCut(1*keV),
   fBetaEKinEndPoint(7.025*MeV),
   fBetaEKinMin(fBetaLowCut),
   fBetaEKinMax(fBetaEKinEndPoint),
   fPosCentre(G4ThreeVector(0,0,0))   
{
    // create messenger 
    primGen_msg = new VKPrimaryGeneratorMessenger(this);

    // General particle source
    fGPSource = new G4GeneralParticleSource();
    fGPSource -> SetCurrentSourceIntensity(1);
  
    // Table of particles
    G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
    G4ParticleDefinition * electron = particleTable -> FindParticle("e-");

    // electron is default particle 
    fGPSource -> GetCurrentSource() -> SetParticleDefinition(electron);

    // Default energy, position, momentum
    fGPSource -> GetCurrentSource() -> GetPosDist() 
              -> SetCentreCoords(G4ThreeVector(0.0 , 0.0, -30.0*cm));
    fGPSource -> GetCurrentSource() -> GetAngDist() 
              -> SetParticleMomentumDirection(G4ThreeVector(0., 0., 1.));
    fGPSource -> GetCurrentSource() -> GetEneDist() 
              -> SetMonoEnergy(1.*MeV);          

    // Initialize radial distribution (for sampling initial position)
    InitRadialDist();
    
    // Get orthogonal vectors perpendicular to fSourceNormal
    GetOrthogonalVectors(fSourceNormal, fOrtho1, fOrtho2);
    
    // default shape function
    shapeFuncCoeff[0] = 1;
    for (int i = 1; i < 6; i++) shapeFuncCoeff[i] = 0;
}


VKPrimaryGeneratorAction::~VKPrimaryGeneratorAction()
{
    delete fGPSource;
    delete fCRMuonZenithDist;
    delete fCRMuonAzimuthalDist;
    delete fRadialDist;
}


void VKPrimaryGeneratorAction::GeneratePrimaries(G4Event* anEvent)
{
    // Is this the first event of the run?
    if (anEvent -> GetEventID() == 0) {
        // if so, store centre position
        fPosCentre = fGPSource -> GetCurrentSource() -> GetPosDist() -> GetCentreCoords(); 
        // and initialize beta spectrum
        if (fBetaSpectrumOn) InitBetaSpectrum();
    }

    // Sample (relative) position coordinates if needed
    if (fDiskShapedSourceOn) 
    {
        // Centre position
        
        // polar coordinates in plane
        G4double angle  = G4UniformRand() * 2*CLHEP::pi *rad;
        G4double radius = fRadialDist -> GetRandom() *cm;
        // calculate position
        fPos = fPosCentre + radius*sin(angle)*fOrtho1 + radius*cos(angle)*fOrtho2;
        fGPSource -> GetCurrentSource() -> GetPosDist() -> SetCentreCoords(fPos);
    }

    // Cosmic-ray muon angular distributions
    if (fCRMuonAngDistOn) 
    {
        // sample zenith angle
        double zen = fCRMuonZenithDist -> GetRandom(fZenithMin/rad, fZenithMax/rad);

        // sample azimuthal angle
        double phi = fCRMuonAzimuthalDist -> GetRandom();
        
        // set direction
        double cz = cos(zen);
        double sz = sin(zen);
        double cp = cos(phi);
        double sp = sin(phi);
        fDirection = G4ThreeVector(sz*cp, -cz, sz*sp);
        fGPSource -> GetCurrentSource() -> GetAngDist() 
                                                -> SetParticleMomentumDirection(fDirection);
    }
    
    // Beta decay energy spectrum
    if (fBetaSpectrumOn) 
    {
        // sample energy
        double emin = fBetaEKinMin / MeV;
        double emax = fBetaEKinMax / MeV;
        G4double ekin = fBetaSpectrumFunc->GetRandom(emin, emax) * MeV;
        fGPSource -> GetCurrentSource() -> GetEneDist() -> SetMonoEnergy(ekin);   
    }
    
    // generate primary
    fGPSource -> GeneratePrimaryVertex(anEvent);
}


void VKPrimaryGeneratorAction::InitCRMuonAngDist()
{
    // functional form
    TString function = Form("sin(x)*pow(cos(x), %f)", fPower);
    // print info
//    G4cout.precision(1);
    G4cout << " ---> Initializing angular distributions for CR muons:" << G4endl;
    G4cout << " ---> f(x) = " << function;
    G4cout << ",  " << fZenithMin/deg << " deg <= x <= " << fZenithMax/deg << " deg" << G4endl;
    // initialize TF1s
    fCRMuonZenithDist    = new TF1("zenithDist", function, 0., CLHEP::pi/2.);
    fCRMuonAzimuthalDist = new TF1("azimuthalDist", "1",   0., 2*CLHEP::pi);
}


void VKPrimaryGeneratorAction::InitRadialDist()
{
    double rmin = 0.5 * fSourceInnerDiameter /cm;
    double rmax = 0.5 * fSourceOuterDiameter /cm;
    fRadialDist = new TF1("radialDist", "x", rmin, rmax);
}



void VKPrimaryGeneratorAction::SetCRMuonAngDistPower(float power) 
{
    fPower = power;
    InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::SetCRMuonAngDistOn(G4bool b) 
{ 
    fCRMuonAngDistOn = b; 
    if (fCRMuonAngDistOn) InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::SetCRMuonMinZenithAngle(G4double angle) 
{
    fZenithMin = angle;
    InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::SetCRMuonMaxZenithAngle(G4double angle) 
{
    fZenithMax = angle;
    InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::GetOrthogonalVectors(const G4ThreeVector& n, G4ThreeVector& o1, G4ThreeVector& o2)
{
    o1 = n.cross(G4ThreeVector(1,0,0));
    if (o1.mag()==0) o1 = n.cross(G4ThreeVector(0,1,0));
    o1 = o1.unit();
    o2 = n.cross(o1);
}


void VKPrimaryGeneratorAction::SetDiskShapedSourceOn(G4bool b) 
{ 
    fDiskShapedSourceOn = b; 
}


void VKPrimaryGeneratorAction::SetSourceInnerDiameter(G4double d)
{
    G4cout << " ---> Source inner diameter set to " << d/mm << " mm" << G4endl;
    fSourceInnerDiameter = d;
    InitRadialDist();
}


void VKPrimaryGeneratorAction::SetSourceOuterDiameter(G4double d)
{
    G4cout << " ---> Source outer diameter set to " << d/mm << " mm" << G4endl;
    fSourceOuterDiameter = d;
    InitRadialDist();
}


void VKPrimaryGeneratorAction::SetSourceNormal(const G4ThreeVector& n)
{
    fSourceNormal = n;
    GetOrthogonalVectors(fSourceNormal, fOrtho1, fOrtho2);
}


// beta spectrum allowed shape
void VKPrimaryGeneratorAction::InitBetaSpectrum()
{
    // allowed shape with F=1 (old)
//    fBetaSpectrumFunc = new TF1("betaSpectrum", "(x+0.511)*sqrt(pow(x+0.511,2)-pow(0.511,2))*pow([0]-x,2)", fBetaLowCut/MeV, fBetaEKinEndPoint/MeV);
//    fBetaSpectrumFunc -> SetParameter(0, fBetaEKinEndPoint/MeV);

    // shape function
    // S(E) = a1 + a2/E + a3 E + a4 E^2 + a5 E^3 + a6 E^4
    shapeFunc = new TF1("shapeFunc", Form("%f + %f/x + %f*x + %f*pow(x,2) + %f*pow(x,3) + %f*pow(x,4)", shapeFuncCoeff[0], shapeFuncCoeff[1], shapeFuncCoeff[2], shapeFuncCoeff[3], shapeFuncCoeff[4], shapeFuncCoeff[5]), 0.5, 1E3);

    // allowed shape with correct Fermi function
    fBetaSpectrum = new VKBetaSpectrum(6, 12, fBetaEKinEndPoint, shapeFunc);
    int npar = 0;
    fBetaSpectrumFunc = new TF1("betaSpectrum", fBetaSpectrum, &VKBetaSpectrum::Evaluate, fBetaLowCut/MeV, fBetaEKinEndPoint/MeV, npar);

    // print info
    G4cout << " ---> Initializing beta-decay energy spectrum (allowed shape) with";
    G4cout << " end-point energy of " << fBetaEKinEndPoint / MeV << " MeV" << G4endl;
}


void VKPrimaryGeneratorAction::SetBetaSpectrumOn(G4bool b) 
{ 
    fBetaSpectrumOn = b; 
    if (fBetaSpectrumOn) InitBetaSpectrum();
}


void VKPrimaryGeneratorAction::SetBetaEndPointEnergy(G4double endPoint) 
{ 
    fBetaEKinEndPoint = endPoint;
    InitBetaSpectrum();
    // if max sampling energy exceeds new end-point energy, reduce it
    if (fBetaEKinMax > endPoint) SetBetaMaxEnergy(endPoint);
}

void VKPrimaryGeneratorAction::SetBetaMinEnergy(G4double emin)      
{     
    fBetaEKinMin = emin;
    
    if (fBetaEKinMin < fBetaLowCut) fBetaEKinMin = fBetaLowCut;
}
        
void VKPrimaryGeneratorAction::SetBetaMaxEnergy(G4double emax)      
{ 
    fBetaEKinMax = emax;     
}

void VKPrimaryGeneratorAction::SetBetaShapeFuncA1(double a1)
{
    shapeFuncCoeff[0] = a1;
}

void VKPrimaryGeneratorAction::SetBetaShapeFuncA2(double a2)
{
    shapeFuncCoeff[1] = a2;
}

void VKPrimaryGeneratorAction::SetBetaShapeFuncA3(double a3)
{
    shapeFuncCoeff[2] = a3;
}

void VKPrimaryGeneratorAction::SetBetaShapeFuncA4(double a4)
{
    shapeFuncCoeff[3] = a4;
}

void VKPrimaryGeneratorAction::SetBetaShapeFuncA5(double a5)
{
    shapeFuncCoeff[4] = a5;
}

void VKPrimaryGeneratorAction::SetBetaShapeFuncA6(double a6)
{
    shapeFuncCoeff[5] = a6;
}

double VKPrimaryGeneratorAction::GetBetaCoverage() const
{
    double coverage = 1;
    double fullIntegral  = fBetaSpectrumFunc -> Integral(fBetaLowCut/MeV, fBetaEKinEndPoint/MeV);
    double restrIntegral = fBetaSpectrumFunc -> Integral(fBetaEKinMin/MeV, fBetaEKinMax/MeV);
    if (fullIntegral > 0) coverage *= restrIntegral / fullIntegral;
    return coverage;
}


double VKPrimaryGeneratorAction::GetCRMuonCoverage() const
{
    double coverage = 1;
    double fullIntegral  = fCRMuonZenithDist -> Integral(0., CLHEP::pi/2.);
    double restrIntegral = fCRMuonZenithDist -> Integral(fZenithMin/rad, fZenithMax/rad); 
    if (fullIntegral > 0) coverage = restrIntegral / fullIntegral;   
    return coverage;
}


double VKPrimaryGeneratorAction::GetCoverage() const
{
    double coverage = 1;

    if (fBetaSpectrumOn) coverage *= GetBetaCoverage();
    
    if (fCRMuonAngDistOn) {
        coverage *= GetCRMuonCoverage();    
    }
    else {
        // theta distribution
        G4double thMin = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMinTheta();
        G4double thMax = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMaxTheta();
        coverage *= (thMax - thMin) / (180. * deg);

        // phi distribution
        G4double phiMin = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMinPhi();
        G4double phiMax = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMaxPhi();
        coverage *= (phiMax - phiMin) / (360. * deg);
    }

    return coverage;
}


G4String VKPrimaryGeneratorAction::GetParticleName() const 
{
    return fGPSource -> GetCurrentSource() -> GetParticleDefinition() -> GetParticleName();
}


G4double VKPrimaryGeneratorAction::GetSourceArea() const 
{
    G4double A = 0;
    if (fDiskShapedSourceOn) {
        G4double ro = fSourceOuterDiameter/2;
        G4double ri = fSourceInnerDiameter/2;
        A = CLHEP::pi * (pow(ro, 2) - pow(ri, 2));
    }
    return A;
}

